﻿<#
    15-01-2018, RASEG
        Navision Stat Powershell bibliotek
    27-03-2018, RASEG
        Tilføjet funktion til monitorering af dataupgrade/konvertering
        + lidt refakturering
    04-07-2018, RASEG
        Reimplementering af ImportFile funktionen.
        Tilføjet loadingbar og generelt pænere output i Start-UpgradeMonitoring.
        Bedre logging.
        Nye hjælpefunktioner + Refakturering.
    10-10-2018, RASEG
        Tilføjet funktion til indlæsning af rettigheder (Automatisk identificering af tenant med app-skriverettigheder)
        Bedre fejlbeskeder ved import - Angivelse af NS_PS_MODULE version
    23-10-2018, RASEG
        Tilføjet funktion til indlæsning af datafiler fra en VMS-pakke
    IN03321, 22-02-2019, RASEG
        Rettet version til NS9.2.005
        Refakturering + omlægning til automatisk identificering og indlæsning af datafiler vha. type og versionering.
        Versioneringslogik ligger i NS, da der skal tjekkes op i mod NS versionen på databasen.
    05-04-2019, RASEG
        Rettet version til NS9.3.000
        Tilføjet mulighed for at angive "TenantsToProcess" så der kun afvikles på specifikke tenants
        Tilføjet Invoke-NavCodeunitAllTenants
    IN03354, 08-05-2019, RASEG
        Rettet version til NS9.3.001
        Rettelser ifm. forberedning for akkumuleret opgradering
    08-10-2019, RASEG
	Rettet version til NS9.4.000
        Rettet "Get-DatabaseVersion" således den tager højde for "Set-TenantsToProcess"
        Tilføjet muligheden for at indlæse Payment Means Priority
            Ny funktion: ImportPaymentMeansPrio
    18-02-2020, RASEG
        Get-DatabaseVersion kaldes nu eksplicit i de funktioner, der skal bruge $Global:CurrDatabaseVersion.
        Funktionen henter kun databaseversionen såfremt den ikke allerede er hentet. 
        Dette fikser et synkroniseringsproblem der kunne opstå under login, hvis vi har ændringer til tabeller der bliver benyttet under login.
    03-04-2021, RASEG
        Rettet version til NS9.7.001
        Man kan nu angive at anvende "Serial" executionmode ved dataopgradering
        Bedre fejlbehandling ved fejlet opstart af dataopgradering
#>


$VERSION                    = 'NS9.7.003'

$Global:PreviouslyInvoked   = $false
$ShowProgressBar            = $true
$ImportCodeunitID           = '6007163'
$DataFilesFolder            = 'Datafiler'
$ExternalComponentsFolder   = 'Eksterne Komponenter'
$ExCompInstallationFolder   = 'Til Installation (eksterne komponenter)'
$ReceiptFolder              = 'Kvitteringsfiler'

$CurrDatabaseVersionPath    = "$PSScriptRoot\CurrDatabaseVersion_temp.txt"
$Global:CurrDatabaseVersion = ''

$Err                        = @()
$Global:SynkError           = $Global:SynkError
$Global:UpgError            = $Global:UpgError
$Global:ImportError         = $Global:ImportError

function Set-ServerInstance($ServerInstance) {
    $Global:ServerInstance = $ServerInstance
}

function Set-TenantsToProcess($TenantsToProcess) {
    if (($TenantsToProcess -eq '') -or (($TenantsToProcess.Count -gt 0) -eq $false)) {
        Remove-Variable $Global:TenantsToProcess
    } else {
        $Global:TenantsToProcess = $TenantsToProcess
    }
}

function Sync-AllTenants {
    param(
        [ValidateSet("CheckOnly", "ForceSync", "Sync")]
        $Mode
    )

    checkServerInstance

    Write-Host -ForegroundColor Yellow "Synkroniserer tenants..."

    $Global:SynkError = $false
    if ($Err) {
        Clear-Variable $Err
    }

    if ($Mode -ne "CheckOnly") {
        $OriginalPref = $ProgressPreference
        $ProgressPreference = "SilentlyContinue"
    }

    if (!$Global:TenantsToProcess) {
        $Tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance
    } else {
        $Tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance | Where Id -In $Global:TenantsToProcess
    }

    $TenantCount = $Tenants.Count
    foreach ($Tenant in $Tenants) {
        Sync-NAVTenant -ServerInstance $Global:ServerInstance -Tenant $Tenant.Id -Mode $Mode -ea silentlycontinue -ErrorVariable Err -Force
    }

    if ($Mode -ne "CheckOnly") {
        $ProgressPreference = $OriginalPref
        Sleep -Milliseconds 50

        if (!$Global:TenantsToProcess) {
            $TenantsOperational = (Get-NavTenant -ServerInstance $Global:ServerInstance | Where {($_.State -ne "OperationalWithSyncInProgress") -and ($_.State -ne "OperationalSyncInProgress") -and ($_.State -ne "Mounted")}).Count
        } else {
            $TenantsOperational = (Get-NavTenant -ServerInstance $Global:ServerInstance | Where Id -In $Global:TenantsToProcess | Where {($_.State -ne "OperationalWithSyncInProgress") -and ($_.State -ne "OperationalSyncInProgress") -and ($_.State -ne "Mounted")}).Count
        }

        $CountLoop = 0
        $Run       = $true
        While (($TenantCount -ne $TenantsOperational) -and ($Run)) {
            if (!$Global:TenantsToProcess) {
                $TenantsOperational = (Get-NavTenant -ServerInstance $Global:ServerInstance | Where {($_.State -ne "OperationalWithSyncInProgress") -and ($_.State -ne "OperationalSyncInProgress") -and ($_.State -ne "Mounted")}).Count
            } else {
                $TenantsOperational = (Get-NavTenant -ServerInstance $Global:ServerInstance | Where Id -In $Global:TenantsToProcess | Where {($_.State -ne "OperationalWithSyncInProgress") -and ($_.State -ne "OperationalSyncInProgress") -and ($_.State -ne "Mounted")}).Count
            }

            if ($CountLoop -gt 480) { #If it continues for more than 8 minutes: STOP
                $Run = $false
                write-host -ForegroundColor Red "`r`nSync-AllTenants ended without all tenants being in an 'Operational' state.`r`nThey may still be syncronizing but an error may also have occured."
            }

            $Percentage = $TenantsOperational/$TenantCount * 100
            $PercentageText = [Math]::Round($Percentage, 2)
            Write-Progress -Id 1 -Activity "Synkroniserer tenants.." -Status "$PercentageText%" -PercentComplete $PercentageText

            $CountLoop++
            Start-Sleep -Seconds 1
        }

        Write-Progress -Id 1 -Activity "Synkroniserer tenants.." -Status Ready -Completed
    }

    if ($Err.Count -gt 0) {
        $Global:SynkError = $true
        write-error $Err[0]  
    } else {
        write-host Synkronisering færdig.
    }

    [Environment]::NewLine   
}

function Invoke-NavCodeunitAllTenants {
    param(
        $CodeunitID,
        $CodeunitMethod,
        $CodeunitArgument
    )

    checkServerInstance

    if (!$Global:TenantsToProcess) {
        $Tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance
    } else {
        $Tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance | Where Id -In $Global:TenantsToProcess
    }

    foreach ($Tenant in $Tenants) {
        [array] $comps = Get-NAVCompany -ServerInstance $ServerInstance -Tenant $Tenant.id
        if ($comps.Count -gt 0) {
            write-host "Afvikler codeunit på tenant:"($tenant.id)
        }

        foreach ($comp in $comps) {                     
            Write-Host "    Regnskab:"$comp.CompanyName 

            if ($CodeunitMethod) {
                if ($CodeunitArgument) {
                    Invoke-NAVCodeunit -Tenant $Tenant.Id -ServerInstance $ServerInstance -codeunitId $CodeunitID -Language 'DA-dk' -CompanyName $comp.CompanyName -MethodName $CodeunitMethod -Argument $CodeunitArgument -ea silentlycontinue -ErrorVariable Err
                } else {
                    Invoke-NAVCodeunit -Tenant $Tenant.Id -ServerInstance $ServerInstance -codeunitId $CodeunitID -Language 'DA-dk' -CompanyName $comp.CompanyName -MethodName $CodeunitMethod -ea silentlycontinue -ErrorVariable Err
                }
            } else {
                Invoke-NAVCodeunit -Tenant $Tenant.Id -ServerInstance $ServerInstance -codeunitId $CodeunitID -Language 'DA-dk' -CompanyName $comp.CompanyName -ea silentlycontinue -ErrorVariable Err
            }

            if ($Err.Count -eq 0) {
                $msg = 'Codeunit: '+$CodeunitID+' succesfuldt afviklet i regnskab "'+$comp.CompanyName+'" for tenant '+$Tenant.Id
                write-host -ForegroundColor Green "      - Succesfuld afvikling af codeunit $CodeunitID i regnskab:"$comp.CompanyName
            } else {
                if (checkError("Codeunit $CodeunitID")) {
                    $msg = 'Fejl ved afvikling af codeunit: '+$CodeunitID+' i regnskab "'+$comp.CompanyName+'" for tenant '+$Tenant.Id
                    write-host -ForegroundColor Red "      - Fejl ved afvikling af codeunit $CodeunitID i regnskab:"$comp.CompanyName
                }
            }

	    write-ToLog -text $msg  

        }
    }
}

function Start-UpgradeMonitoring {
    param(
        [ValidateSet("Parallel", "Serial")]
        $ExecutionMode,
        [switch] $InvokeUpgrade
    )

    if ($Global:SynkError) {
        write-host -ForegroundColor red "Dataopgraderingen kan ikke startes, da der er synkroniseringsfejl!"
        exit
    }

    if(!$ExecutionMode) {
        $ExecutionMode = "Parallel"
    }

    checkServerInstance  
    Get-DatabaseVersion

    if ($InvokeUpgrade) {
        $StartedMsg = "<- Datakonvertering igangsat ($ExecutionMode) ->"
    } else {
        $StartedMsg = "<- Upgrade monitorering igangsat ->"
    }

    Write-Host -NoNewline $StartedMsg
    write-ToLog $StartedMsg

    $stopwatch = [System.Diagnostics.Stopwatch]::StartNew()

    $Global:UpgError = $false
    $upgradeStartError = $false
    $tenantsUpdateList = @()

    if (!$Global:TenantsToProcess) {
        $tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance
    } else {
        $tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance | Where Id -In $Global:TenantsToProcess
    }
    
    #Collecting tenants and starting upgrade if requested
    foreach($tenant in $tenants) { 
        $tenantsUpdateList += ,($tenant.Id, $false)
          
        if ($InvokeUpgrade) { 
            Start-NAVDataUpgrade -ServerInstance $Global:ServerInstance -Tenant $tenant.Id -Language 'da-DK' -FunctionExecutionMode $ExecutionMode -ErrorAction SilentlyContinue -ErrorVariable Err -Force
               
            if (!$?) {
                $upgradeStartError = $true
            }
        }
    }

    if ($upgradeStartError) {     
        $ErrMsg = ""
        try {
            $ErrMsg = $Err[0].ToString()
        } catch {}

        try {
            $ErrMsg += $error[0].InvocationInfo.PositionMessage
        } catch {}

	    if ($ErrMsg -ne "") {
        	    if ($ErrMsg.Contains("kan ikke startes i den aktuelle lejertilstand:")`
       		    -or ($ErrMsg.Contains("cannot be started in the current tenant state:"))) {
            		    $Global:SynkError = $true
            		    Write-host "En eller flere tenants mangler at blive synkroniseret (Dataopgraderingen fejlede opstart ved følgende tenant: $($tenant.Id)). Synkroniser venligst tenants igen med mode: CheckOnly og kontrollér for fejl. - Monitorering afbrudt." -ForegroundColor Yellow
        	    } else {
            		    Write-Host "Kontrollér fejlen: $ErrMsg - Dataopgradering ikke startet for alle tenants - Monitorering afbrudt." -ForegroundColor Yellow
        	    }
	    } else {
            Write-Error "Dataopgraderingen kunne ikke startes! (Dataopgraderingen fejlede opstart ved følgende tenant: $($tenant.Id)) Monitorering afbrudt."
        }

        $Global:UpgError = $true

        exit;
    }

    #Variables for total percentage calculation
    $tc_count = 0
    $tc_total = 10000 * $tenants.Count

    #Continously update
    $stop = $false
    $total = $tenants.Count
    $rounds = 0
    $doneCount = 0
    $maxRounds = 2000
    $updateFreq = 750 #Milliseconds
    $lineBreak = 4
    while (!$stop){
        
        $tc_count = 0
        $lineCount = 0

        foreach($tenant in $tenantsUpdateList) {       
            $doneCountIncremented = $false

            $NAVDataUpgradeStatus = Get-NAVDataUpgrade -ServerInstance $Global:ServerInstance -Tenant $tenant[0]

            $color = 'Gray'
            [decimal]$returnedDecimal = 10000
            [decimal]::TryParse($NAVDataUpgradeStatus.Progress.Substring(0, $NAVDataUpgradeStatus.Progress.Length-2), [ref]$returnedDecimal) | Out-Null
            $tc_count += $returnedDecimal

            if (($NAVDataUpgradeStatus.State -eq 'Stopped') -or ($NAVDataUpgradeStatus.State -eq 'Completed')) {
                if (!$tenant[1]) {
                    $doneCount += 1
                    $tenant[1] = $true
                }
            }
        
            if (!$tenant[1]) {
                $ErrorsDataUpgrade = Get-NAVDataUpgrade -ServerInstance $Global:ServerInstance -Tenant $tenant[0] -ErrorOnly
                if ($ErrorsDataUpgrade) {
                    foreach($ErrorDataUpgrade in $ErrorsDataUpgrade){
                        Write-Host "Fejl i funktion $($ErrorDataUpgrade.FunctionName) for regnskab $($ErrorDataUpgrade.CompanyName)`r`n $($ErrorDataUpgrade.Error)" -ForegroundColor red
                        $Global:UpgError = $true
                    }
                    $doneCount += 1
                    $tenant[1] = $true
                }
            }
        }

        if ($doneCount -eq $total) {
            $stop = $true
        }

        #Security stop if endless loop
        if ($rounds -eq $maxRounds) {
            $stop = $true
            Write-Host "Datakonvertering tager meget lang tid - monitorering afbrudt. Brug 'Get-NavDataUpgrade' for at følge processen" -ForegroundColor yellow
        }

        $tc_percentage = $tc_count/$tc_total * 100
        $tc_percentageText = [Math]::Round($tc_percentage, 2)
        Write-Progress -Id 1 -Activity "Datakonvertering igang.." -Status "$tc_percentageText%" -PercentComplete $tc_percentageText

        $rounds += 1
        Start-Sleep -Milliseconds $updateFreq
    }

    if ($Global:UpgError) {
        write-host -ForegroundColor Red "Dataopgraderingsfejl!"
    }

    Write-Progress -Id 1 -Activity Upgrading.. -Status "$tc_percentageText%" -PercentComplete $tc_percentageText -Completed
    Print-ElapsedTime($stopwatch)
    Write-host "`n<- Datakonvertering afsluttet ->"
    write-ToLog "<- Datakonvertering afsluttet ->"
}

function Import-NSVMS {
    param(
        $serverInstance,
        $package
    )

    if ($Global:SynkError) {
        write-host -ForegroundColor red "Importeringen kan ikke startes, da der er synkroniseringsfejl!"
        exit
    }

    if ($Global:UpgError) {
        write-host -ForegroundColor red "Importeringen kan ikke startes, da der er dataopgraderingsfejl!"
        exit;
    }

    write-host "<- Indlæsning af datafiler i vms pakke $package startet på serverinstans: $Global:ServerInstance ->`n"
    write-ToLog -text "<- Indlæsning af datafiler i vms pakke $package startet på serverinstans: $Global:ServerInstance ->"
        
    $FilePath = "$PSScriptRoot\$File"
        
    $Companies = Get-NAVCompany -ServerInstance $Global:ServerInstance
    
    foreach ($Company in $Companies) {
        Write-Host "Indlæser i regnskab:"$Company.CompanyName"..."
        Invoke-NAVCodeunit -ServerInstance $Global:ServerInstance -codeunitId $ImportCodeunitID -Language 'DA-dk' -CompanyName $Company.CompanyName -MethodName 'VMSImport' -Argument $package -ea silentlycontinue -ErrorVariable Err

        if ($Err.Count -eq 0) {
            $msg = "Succesfuld indlæsning af datafiler i vms pakke $package i regnskab " + $Company.CompanyName
            write-host -ForegroundColor Green "  -Succesfuld indlæsning af datafiler i vms pakke $package i regnskab "$Company.CompanyName
        } else {
            if (checkError("vms pakke $package")) {
                $msg = "Fejl ved indlæsning af datafiler i vms pakke $package" + $Company.CompanyName
                write-host -ForegroundColor Red "  -Fejl ved indlæsning af datafiler i vms pakke $package i regnskab "$Company.CompanyName
            }
        }             
        write-ToLog -text $msg  
    }

    write-ToLog -text "<- Afsluttet indlæsning af datafiler i vms pakke $package ->"
    write-host "`n<- Afsluttet indlæsning af datafiler i vms pakke $package ->`n" 
}

function Import-NSDatafiles() {
    
    if ($Global:SynkError) {
        write-host -ForegroundColor red "Importeringen kan ikke startes, da der er synkroniseringsfejl!"
        exit
    }

    if ($Global:UpgError) {
        write-host -ForegroundColor red "Importeringen kan ikke startes, da der er dataopgraderingsfejl!"
        exit;
    }

    $dataFiles = gci "$PSScriptRoot\$DataFilesFolder" | % {
        $type  = $_.BaseName.Substring(0, 3).ToUpper()
        $file  = $_.Name
        switch($type) {
            "RET" { Import-NSPermissions -File $file break }
            "TSB" { Import-NSXMLBro -File $file break }
            "SKE" { Import-NSKontoskema -File $file break }
            "OIO" { Import-NSBetalingsmetode -File $file break }
            "MOM" { Import-NSMomsAngivelse -File $file break }
            "KTY" { Import-NSEBilagNøgleType -File $file break }
            "KLA" { Import-NSClassification -File $file break }
            "DIM" { Import-NSDimVærdier -File $file break }
            "CMS" { Import-NSCollMgtSetup -File $file break }
            "GTF" { Import-NSGISTabelogFeltmapninger -File $file break }
            "FOR" { Import-NSKolonneFormat -File $file break }
            "DPX" { Import-NSDigitalPostBroOpsætning -File $file break }
            "PRO" { Import-NSTSProfilOpsætning -File $file break }
            "CIS" { Import-NSCISDatastream -File $file break }
            "PMP" { Import-PaymentMeansPrio -File $file break }
            {$_ -in `
            "BA1",  `
            "BA2",  `
            "BA3"}{ Import-NSBankSetup -File $file break }
          Default { 
            Write-Host -ForegroundColor Red "$file, Ukendt type af datafil: $type!" ([Environment]::NewLine) 
            $Global:ImportError = $true
          }
        }
    }

    if ($Global:ImportError) {
        write-host -ForegroundColor Red "Datafiler importeret med kritiske fejl undervejs!"
    } 
}

function Import-NSDimVærdier($File) {                   ##Indkøbskategorier
    importFile -File $File -Type 'ImportDimValues'
}

function Import-NSEBilagNøgleType($File) {
    importFile -File $File -Type 'ImportEDocumentKeyType'
}

function Import-NSMomsAngivelse($File) {
    importFile -File $File -Type 'ImportVatStatement'
}

function Import-NSTSProfilOpsætning($File) {
    importFile -File $File -Type 'ImportNSTSProfile'
}

function Import-NSXMLBro($File) {
    importFile -File $File -Type 'ImportXMLBridge'
}

function Import-NSKolonneFormat($File) {
    importFile -File $File -Type 'ImportColumnLayout'
}

function Import-NSDigitalPostBroOpsætning($File) {
    importFile -File $File -Type 'ImportDPXMLBridge'
}

function Import-NSKontoskema($File) {
    importFile -File $File -Type 'ImportAccShedule'
}

function Import-NSBetalingsmetode($File) {
    importFile -File $File -Type 'ImportPaymentMethod'
}

function Import-NSClassification($File) {
    importFile -File $File -Type 'ImportClassification'
}

function Import-NSBankSetup($File) {
    importFile -File $File -Type 'ImportBankSetup'
}

function Import-NSCollMgtSetup($File) {
    importFile -File $File -Type 'ImportCollMgtSetup'
}

function Import-NSGISTabelogFeltmapninger($File) {
    importFile -File $File -Type 'ImportGisInterfaceSetup'
}

function Import-NSCISDatastream($File) {
    importFile -File $File -Type 'ImportCISDatastream'
}

function Import-PaymentMeansPrio($File) {
    importFile -File $File -Type 'ImportPaymentMeansPrio'
}

function Import-NSPermissions($File) {
    
    if ($Global:SynkError) {
        write-host -ForegroundColor red "Importeringen kan ikke startes, da der er synkroniseringsfejl!"
        exit
    }

    if ($Global:UpgError) {
        write-host -ForegroundColor red "Importeringen kan ikke startes, da der er dataopgraderingsfejl!"
        exit;
    }

    checkServerInstance
    Get-DatabaseVersion
    
    $Global:ImportError = $false
    $AppTenantId = ''
    $Tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance | where "AllowAppDatabaseWrite" -eq "True"
    if ($Tenants.Count -le 0) {
        Write-Error "En tenant med skriveadgang til applikationsdatabasen kunne ikke findes. Dette er påkrævet for at kunne indlæse rettigheder!"
        $Global:ImportError = $true
    } else {
        if ($Global:TenantsToProcess) {
            $AppTenantFound = $false
            foreach ($ti in $Tenants) {
                if ($ti.Id -in $Global:TenantsToProcess) {
                    $AppTenantFound = $true
                    $AppTenantId = $ti.Id
                }
            }
            if (-not $AppTenantFound) {
                Write-Error "En tenant med skriveadgang til applikationsdatabasen blev fundet - men fremgår ikke af listen med `"Tenants to process`". Rettigheder kunne ikke indlæses!"
                $Global:ImportError = $true
            }
        } else {
            $AppTenantId = $Tenants[0].Id
        }
    }

    if ($AppTenantId -ne '') {
        write-host "<- Indlæsning af rettigheder i '$File' igangsat via tenant: $AppTenantId->`n"
        write-ToLog -text "<- Indlæsning af rettigheder i '$File' igangsat via tenant: $AppTenantId->"
        
        $FilePath = "$PSScriptRoot\$DataFilesFolder\$File"
        
        $Companies = Get-NAVCompany -ServerInstance $Global:ServerInstance -Tenant $AppTenantId
        if ($Companies.Count -gt 1) {
            $Company = $Companies[0].CompanyName
        } else {
            $Company = $Companies.CompanyName
        }

        Invoke-NAVCodeunit -Tenant $AppTenantId -ServerInstance $Global:ServerInstance -codeunitId $ImportCodeunitID -Language 'DA-dk' -CompanyName $Company -MethodName 'ImportPermissions' -Argument "$Global:CurrDatabaseVersion#$FilePath" -ea silentlycontinue -ErrorVariable Err

        if ($Err.Count -eq 0) {
            $msg = 'Rettigheder i filen: '+$File+' indlæst via tenant '+$AppTenantId
            write-host -ForegroundColor Green "Succesfuld indlæsning af rettigheder i '$File' indlæst via tenant:"$AppTenantId
        } else {
            if (checkError($File)) {
                $msg = 'Fejl ved indlæsning af rettigheder i filen: '+$File+' forsøgt indlæst via tenant '+$AppTenantId
                write-host -ForegroundColor Red "Fejl ved indlæsning af rettigheder i '$File' via tenant:"$AppTenantId
                $Global:ImportError = $true
            }
        }      
        
        write-ToLog -text $msg  

        write-ToLog -text "<- Indlæsning af rettigheder i '$File' afsluttet ->"
        write-host "`n<- Indlæsning af rettigheder i '$File' afsluttet ->`n" 
    }
}

function importFile {
    param($File,
          $Type
    )

    if ($Global:SynkError) {
        write-host -ForegroundColor red "Importeringen kan ikke startes, da der er synkroniseringsfejl!"
        exit
    }

    if ($Global:UpgError) {
        write-host -ForegroundColor red "Importeringen kan ikke startes, da der er dataopgraderingsfejl!"
        exit;
    }

    checkServerInstance
    Get-DatabaseVersion

    if (!$File) {
        write-host -ForegroundColor Red ANGIV 'Fil' !
        exit
    } else {
        $stopwatch = [System.Diagnostics.Stopwatch]::StartNew()

        if (-not $Global:PreviouslyInvoked) {
            [Environment]::NewLine
            [Environment]::NewLine
        }

        $Global:ImportError = $false
        $FilePath = "$PSScriptRoot\$DataFilesFolder\$File"
        $TotalJobs = 0
        $TotalComps = 0
        $first = $true
        $measureCompArray = @()

        if (!$Global:TenantsToProcess) {
            $Tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance
        } else {
            $Tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance | Where Id -In $Global:TenantsToProcess
        }

        foreach ($Tenant in $Tenants) {
            [array] $comps = Get-NAVCompany -ServerInstance $Global:ServerInstance -Tenant $tenant.id
            $TotalComps += $comps.Count
        }

        write-host "<- Indlæsning af $File i $TotalComps regnskaber igangsat ->`n"
        write-ToLog "<- Indlæsning af $File i $TotalComps regnskaber igangsat ->"

        foreach ($Tenant in $Tenants) {
            [array] $comps = Get-NAVCompany -ServerInstance $Global:ServerInstance -Tenant $tenant.id
            if ($comps.Count -gt 0) {
                write-host "Importerer"$File" for tenant:"($tenant.id)
            }
            foreach ($comp in $comps) {
                if (-not $first) {
                    $measureCompArray += [Math]::Round($measureComp.Elapsed.TotalSeconds)

                    $secSum = 0
                    foreach($s in $measureCompArray) {
                        $secSum += $s
                    }

                    $avgSec = $secSum / $measureCompArray.Count

                    $minutes = ($avgSec * ($totalComps - $totalJobs)) / 60
                    $calcLeft = "- Estimeret tilbageværende tid: $minutes minutter"
                    if ($minutes -lt 1) {
                        $minutes = ($avgSec * ($totalComps - $totalJobs))
                        $calcLeft = "- Estimeret tilbageværende tid: $minutes sekunder"
                    }
                } else {
                    $calcLeft = "- Estimeret tilbageværende tid: beregner.."
                }
                
                $measureComp = [System.Diagnostics.Stopwatch]::StartNew()
                
                Write-Host "    Regnskab:"$comp.CompanyName 
                update-ImportProgressBar -TotalComps $TotalComps -TotalJobs $TotalJobs -CalcLeft $calcLeft
                Invoke-NAVCodeunit -Tenant $Tenant.Id -ServerInstance $Global:ServerInstance -codeunitId $ImportCodeunitID -Language 'DA-dk' -CompanyName $comp.CompanyName -MethodName $Type -Argument "$Global:CurrDatabaseVersion#$FilePath" -ea silentlycontinue -ErrorVariable Err

                if ($Err.Count -eq 0) {
                    $msg = 'Filen: '+$File+' importeret i regnskab "'+$comp.CompanyName+'" for tenant '+$Tenant.Id
                    write-host -ForegroundColor Green "      - Succesfuld import af $File i regnskab:"$comp.CompanyName
                } else {
                    if (checkError("$File")) {
                        $msg = 'Fejl ved importering af: '+$File+' i regnskab "'+$comp.CompanyName+'" for tenant '+$Tenant.Id
                        write-host -ForegroundColor Red "      - Fejl ved import af $File i regnskab:"$comp.CompanyName
                        $Global:ImportError = $true
                    }
                }

                $measureComp.stop()
                $first = $false
                write-ToLog $msg
                $TotalJobs++
            }
        }

        Write-Progress -Id 1 -Activity "Indlæser $File i $TotalComps regnskaber.." -PercentComplete 100 -Completed 
        Print-ElapsedTime($stopwatch)
        write-ToLog -text "<- Indlæsning af $File afsluttet ->"
        write-host "`n<- Indlæsning af $File afsluttet ->`n" 

        $Global:PreviouslyInvoked = $true
    }    
}

function Get-ExternalComponents {
    if ($Global:SynkError) {
        write-host -ForegroundColor red "Eksterne komponenter forberedes ikke, da der er synkroniseringsfejl!"
        exit
    }

    if ($Global:UpgError) {
        write-host -ForegroundColor red "Eksterne komponenter forberedes ikke, da der er dataopgraderingsfejl!"
        exit;
    }

    if ($Global:ImportError) {
        write-host -ForegroundColor red "Eksterne komponenter forberedes ikke, da der er importeringsfejl!"
        exit;
    }
    
    checkServerInstance
    Get-DatabaseVersion

    $ToBeInstalled = @()

    $Components = gci "$PSScriptRoot\$ExternalComponentsFolder" | % {
        $Version          = [int] $_.BaseName.Split('_')[0]
        $CurrVersionCheck = [int] $Global:CurrDatabaseVersion.Replace('.', '')
        
        if ($version -gt $CurrVersionCheck) {
            $ToBeInstalled += $_.FullName
        }
    }

    if ($ToBeInstalled.Length -gt 0) {
        if (!(Test-Path "$PSScriptRoot/$ExCompInstallationFolder")) {
            new-item "$PSScriptRoot/$ExCompInstallationFolder" -ItemType Directory | Out-Null
        } else {
            Remove-Item "$PSScriptRoot/$ExCompInstallationFolder" -Recurse -force | Out-Null
            new-item "$PSScriptRoot/$ExCompInstallationFolder" -ItemType Directory | Out-Null
        }
        
        [Environment]::NewLine
        write-host -ForegroundColor Yellow "Følgende komponenter skal installeres manuelt: "

        foreach($C in $ToBeInstalled) {
            Copy-Item $C "$PSScriptRoot/$ExCompInstallationFolder" 

            $GetNameSplit = $C.Split('\')
            $FileName     = $GetNameSplit[$GetNameSplit.length - 1]

            write-host "    - $PSScriptRoot\$ExCompInstallationFolder\$FileName"
        }
    }
}

function Export-Receipt {
    if ($Global:SynkError) {
        write-host -ForegroundColor red "Kvitteringsfil kan ikke udlæses, da der er synkroniseringsfejl!"
        exit
    }

    if ($Global:UpgError) {
        write-host -ForegroundColor red "Kvitteringsfil kan ikke udlæses, da der er dataopgraderingsfejl!"
        exit;
    }

    if ($Global:ImportError) {
        write-host -ForegroundColor red "Kvitteringsfil kan ikke udlæses, da der er importeringsfejl!"
        exit;
    }
    
    CheckServerInstance

    [Environment]::NewLine
    Write-host -ForegroundColor Yellow "Udlæser kvitteringsfiler..."

    if (!$Global:TenantsToProcess) {
        $Tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance
    } else {
        $Tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance | Where Id -In $Global:TenantsToProcess
    }

    if ($Tenants.Count -gt 0) {
        if (!(Test-Path "$PSScriptRoot/$ReceiptFolder")) {
            new-item "$PSScriptRoot/$ReceiptFolder" -ItemType Directory | Out-Null
        } else {
            Remove-Item "$PSScriptRoot/$ReceiptFolder" -Recurse -force | Out-Null
            new-item "$PSScriptRoot/$ReceiptFolder" -ItemType Directory | Out-Null
        }
    }

    $Path = "$PSScriptRoot\$ReceiptFolder"

    foreach($Tenant in $Tenants) {
        $Companies = Get-NAVCompany -ServerInstance $Global:ServerInstance -Tenant $Tenant.Id
        if ($Companies.Count -gt 1) {
            $Company = $Companies[0].CompanyName
        } else {
            $Company = $Companies.CompanyName
        }

        Invoke-NAVCodeunit -Tenant $Tenant.Id -ServerInstance $Global:ServerInstance -codeunitId $ImportCodeunitID -CompanyName $Company -Language 'DA-dk' -MethodName ExportReceipt -Argument $Path
    }

    Write-host "Kvitteringsfiler udlæst"
}

function Export-UPGTenantList {
    $outputFilePath = "$PSScriptRoot\UPG Tenant List.txt"

    if (!$Global:TenantsToProcess) {
        $Tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance
    } else {
        $Tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance | Where Id -In $Global:TenantsToProcess
    }

    "Id; DatabaseName; DatabaseServer; AllowAppDatabaseWrite; NasServicesEnabled; RunWithAdminRights; DefaultCompany; DefaultTimeZone" | Out-File $outputFilePath -Encoding utf8

    foreach ($tenant in $Tenants) {
        $tenantCsvInfoLine = $tenant.Id
        $tenantCsvInfoLine += "; $($tenant.DatabaseName)"
        $tenantCsvInfoLine += "; $($tenant.DatabaseServer)"
        $tenantCsvInfoLine += "; $($tenant.DatabaseInstance)"
        $tenantCsvInfoLine += "; $($tenant.AllowAppDatabaseWrite)"
        $tenantCsvInfoLine += "; $($tenant.NasServicesEnabled)"
        $tenantCsvInfoLine += "; $($tenant.RunNasWithAdminRights)"
        $tenantCsvInfoLine += "; $($tenant.DefaultCompany)"
        $tenantCsvInfoLine += "; $($tenant.DefaultTimeZone)"

        $tenantCsvInfoLine | Out-File $outputFilePath -Encoding utf8 -Append
    }
}

function checkError($info) {
    if ($Err[0].ToString().contains("med tilsvarende argumenttyper.")) {
        $eMsg = "Codeunit $ImportCodeunitID ImportNSDataFiles er ikke den korrekte version. NS_PS_MODULE kører version: $VERSION"
        write-host -ForegroundColor Red $eMsg
        write-ToLog $eMsg
        return $false
    } elseif($Err[0].ToString().contains("ALREADY IMPORTED")) {
        $eMsg = "$info er allerede importeret"
        write-host -ForegroundColor yellow "      -"$eMsg
        write-ToLog $eMsg
        return $false
    } elseif($Err[0].ToString().contains("benytter sig umiddelbart ikke af CM")) {
        $eMsg = $Err[0].ToString()
        write-host -ForegroundColor Red "      -"$eMsg
        write-ToLog $eMsg
        return $false
    } else {
        $eMsg = "NS_PS_MODULE VERSION $VERSION $([Environment]::NewLine)$info$([Environment]::NewLine)" + $Err[0].ToString()
        write-host -ForegroundColor Red $eMsg
        write-ToLog $eMsg
        return $true
    }
}

function Get-DatabaseVersion {
    if ($Global:CurrDatabaseVersion -ne '') {
        return
    }
    
    if (!$Global:TenantsToProcess) {
        $Tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance
    } else {
        $Tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance | Where Id -In $Global:TenantsToProcess
    }
    
    if ($Tenants.Count -gt 1) {
        $Tenant = $Tenants[0].Id
    } else {
        $Tenant = $Tenants.Id
    }

    $Companies = Get-NAVCompany -ServerInstance $Global:ServerInstance -Tenant $Tenant
    if ($Companies.Count -gt 1) {
        $Company = $Companies[0].CompanyName
    } else {
        $Company = $Companies.CompanyName
    }

    Invoke-NAVCodeunit -Tenant $Tenant -ServerInstance $Global:ServerInstance -codeunitId $ImportCodeunitID -CompanyName $Company -Language 'DA-dk' -MethodName SaveDatabaseVersion -Argument $CurrDatabaseVersionPath
    
    $Global:CurrDatabaseVersion = Get-Content $CurrDatabaseVersionPath
    Remove-Item $CurrDatabaseVersionPath
}

function checkServerInstance {
    if (!$Global:ServerInstance) {
        write-host -ForegroundColor Red ANGIV 'Serverinstance!'
        exit
    }
}

function update-ImportProgressBar {
    param(
        $TotalComps,
        $TotalJobs,
	    $CalcLeft
    )
    if ($ShowProgressBar) {
        Write-Progress -Id 1 -Activity "Indlæser $File i $TotalComps regnskaber $CalcLeft" -PercentComplete ($TotalJobs / $TotalComps * 100) -Status "$($TotalComps - $TotalJobs) tilbage.." 
    }
}

function write-ToLog([parameter(ValueFromPipeline)] $text) {
    if (!$text) {
        return
    }
    
    $logTime = Get-Date -Format "hh:mm:ss"
    $path = $PSScriptRoot + "\NSPowerShellLog " + (get-date -format "M-d-yyyy") + '.txt'
    if (-not (Test-Path $path)) {
        New-Item -ItemType file -Path $path | Out-Null
    }
    $line = $logTime + ": " + $text
    $line >> $Path
}

function Print-ElapsedTime($stopwatch) {
    [Environment]::NewLine
    write-host Total tid for afvikling:
    $stopwatch.stop()

    $elapsedText = "$($stopwatch.elapsed.Seconds) sekunder"
    if ($stopwatch.Elapsed.Minutes -gt 0) {$elapsedText = "$($stopwatch.Elapsed.Minutes) minutter, $elapsedText"} 
    if ($stopwatch.Elapsed.Hours -gt 0) {$elapsedText = "$($stopwatch.Elapsed.Hours) timer, $elapsedText"} 
    if ($stopwatch.Elapsed.Days -gt 0) {$elapsedText = "$($stopwatch.Elapsed.Days) dage, $elapsedText"} 

    Write-host $elapsedText
    write-ToLog "Total tid for afvikling: $elapsedText"
}

function Invoke-UpgradePerTenant {
    checkServerInstance

    $Tenants = Get-NAVTenant -ServerInstance $Global:ServerInstance
    foreach ($Tenant in $Tenants) {
        Start-NAVDataUpgrade -ServerInstance $Global:ServerInstance -Tenant $Tenant.Id -Language 'da-DK' -FunctionExecutionMode Parallel -Force
    }
    [Environment]::NewLine   
}

##MS ADMIN MODULE START##
#--------------------------------------------------------------------------
#     Copyright (c) Microsoft Corporation.  All rights reserved.
#--------------------------------------------------------------------------
$errorVariable = $null

# Import-Module or register Snap-in, that will enable side-by-side registrations
function RegisterSnapIn($snapIn, $visibleName)
{
  $nstPath = "HKLM:\SOFTWARE\Microsoft\Microsoft Dynamics NAV\90\Service"

  $snapInAssembly = Join-Path (Get-ItemProperty -path $nstPath).Path "\$snapIn.psm1"
  if(!(Test-Path $snapInAssembly)) { $snapInAssembly = Join-Path (Get-ItemProperty -path $nstPath).Path "\$snapIn.psd1" }
  if(!(Test-Path $snapInAssembly)) { $snapInAssembly = Join-Path (Get-ItemProperty -path $nstPath).Path "\$snapIn.dll" }

  # First try to import the Snap-in
  Import-Module $snapInAssembly -ErrorVariable errorVariable -ErrorAction SilentlyContinue
  
  if (Check-ErrorVariable -eq $true)
  {
    # fallback to add the snap-in
    if ((Get-PSSnapin -Name $snapIn -ErrorAction SilentlyContinue) -eq $null)
    {
        if ((Get-PSSnapin -Registered $snapIn -ErrorAction SilentlyContinue) -eq $null)
        {
	    <# RASEG: UDKOMMENTERET
            write-host -fore Red "Couldn't register $visibleName"
            write-host -fore Red "Some cmdlets may not be available`n"
	    #>
        }
        else
        {
            Add-PSSnapin $snapIn            
        }
    }
  }
}

# Check if there is any error in the ErrorVariable
function Check-ErrorVariable
{
    return ($errorVariable -ne $null -and $errorVariable.Count -gt 0)
}

# Register Microsoft Dynamics NAV Management Snap-in
RegisterSnapIn "Microsoft.Dynamics.Nav.Management" "Microsoft Dynamics NAV Management Snap-in"

# Register Microsoft Dynamics NAV Apps Management Snap-in
RegisterSnapIn "Microsoft.Dynamics.Nav.Apps.Management" "Microsoft Dynamics NAV App Management Snap-in"


#Export-ModuleMember -Function Invoke-UpgradePerTenant

Export-ModuleMember -Function Invoke-NavCodeunitAllTenants
Export-ModuleMember -Function Sync-AllTenants
Export-ModuleMember -Function Start-UpgradeMonitoring
Export-ModuleMember -Function Set-ServerInstance
Export-ModuleMember -Function Set-TenantsToProcess
Export-ModuleMember -Function Get-DatabaseVersion
Export-ModuleMember -Function Get-ExternalComponents
Export-ModuleMember -Function Import-NSVMS
Export-ModuleMember -Function Import-NSDatafiles
Export-ModuleMember -Function Import-NSDimVærdier
Export-ModuleMember -Function Import-NSEBilagNøgleType
Export-ModuleMember -Function Import-NSMomsAngivelse
Export-ModuleMember -Function Import-NSTSProfilOpsætning
Export-ModuleMember -Function Import-NSXMLBro
Export-ModuleMember -Function Import-NSKolonneFormat
Export-ModuleMember -Function Import-NSDigitalPostBroOpsætning
Export-ModuleMember -Function Import-NSKontoskema
Export-ModuleMember -Function Import-NSBetalingsmetode
Export-ModuleMember -Function Import-NSClassification
Export-ModuleMember -Function Import-NSBankSetup
Export-ModuleMember -Function Import-NSCollMgtSetup
Export-ModuleMember -Function Import-NSGISTabelogFeltmapninger
Export-ModuleMember -Function Import-NSCISDatastream
Export-ModuleMember -Function Import-NSPermissions
Export-ModuleMember -Function Import-PaymentMeansPrio
Export-ModuleMember -Function Export-Receipt
Export-ModuleMember -Function Export-UPGTenantList
